#ifndef DCE_MANAGER_HELPER_H
#define DCE_MANAGER_HELPER_H

#include "ns3/object-factory.h"
#include "ns3/attribute.h"
#include "ns3/object-base.h"
#include "ns3/node-container.h"
#include "ns3/nstime.h"
#include <string>

namespace ns3 {

class DceManagerHelper : public Object
{
public:
  static TypeId GetTypeId (void);
  virtual TypeId GetInstanceTypeId (void) const;
  DceManagerHelper ();
  void SetScheduler (std::string type, 
		     std::string n0 = "", const AttributeValue &v0 = EmptyAttributeValue (),
		     std::string n1 = "", const AttributeValue &v1 = EmptyAttributeValue ());
  void SetTaskManagerAttribute (std::string n0, const AttributeValue &v0);
  void SetLoader (std::string type);
  void SetNetworkStack (std::string type,
			std::string n0 = "", const AttributeValue &v0 = EmptyAttributeValue ());
  void SetAttribute (std::string n1, const AttributeValue &v1);
  void Install (NodeContainer nodes);
  // Path used by simulated methods 'execvp' and 'execlp'
  void SetVirtualPath (std::string p);
  std::string GetVirtualPath () const;
private:
  static void RunIp (Ptr<Node> node, Time at, std::string str);
  static void AddAddress (Ptr<Node> node, Time at, std::string name, std::string address);
  void AddRoutes (Ptr<Node> node, std::string r);
  void AddRoute (Ptr<Node> node, std::string r);

  ObjectFactory m_loaderFactory;
  ObjectFactory m_schedulerFactory;
  ObjectFactory m_taskManagerFactory;
  ObjectFactory m_managerFactory;
  ObjectFactory m_networkStackFactory;
  std::string m_virtualPath;
};

} // namespace ns3

#endif /* DCE_MANAGER_HELPER_H */
